/************************************************************
	FontTaffer - Thief 3 font creation program
	by Crispy

"Taffing around with Thief 3 fonts so you don't have to..."

A simple command-line program that takes a TTF or FON
file (TrueType and raster fonts, respectively) and
creates a CEL text file and BMP image file. Once the
BMP file has been converted into a DDS texture file,
Thief 3 will be able to use the font.

Written in the Digital Mars D programming language.
See http://digitalmars.com/d/

************************************************************/

char[] progVersion = "0.6.0";

private {
	import derelict.sdl.sdl;
	import derelict.sdl.ttf;
	import derelict.util.exception;
	import getopt;
	import std.file;
	import std.path;
	import std.stdio;
	import std.string;
}

int abs(int i) {
	if (i<0) return -i;
	else return i;
}

int main(char[][] args) {
	//// Parse arguments
	
	char[] inputfilename="";
	char[] outputfilename="";
	uint fontsize=36;
	uint texturewidth=512, textureheight=512;
	float linespacing = 1.5;
	long fontindex = 0;
	bool fontbold = false, fontitalic = false, fontunderline = false;
	int offset_left = 0, offset_top = 0;
	
	// Take out the program name (args[0]):
	args = args[1..args.length];
	
	// Get input filename
	if (args.length>0 && (args[0][0] != '-')) {
		inputfilename = args[0];
		args = args[1..args.length];
	}
	
	// Get other options
	getopt_t[] usedoptions;
	
	char[] shortoptions = "ho:s:t:x:biu:v:l:";
	char[][11] longoptions;
	longoptions[0] = "help";
	longoptions[2] = "output=";
	longoptions[3] = "size=";
	longoptions[4] = "texture-size=";
	longoptions[5] = "index=";
	longoptions[6] = "bold";
	longoptions[7] = "italic";
	longoptions[8] = "offset-left=";
	longoptions[9] = "offset-top=";
	longoptions[9] = "line-spacing=";
	
	try {
		usedoptions = getopt.getopt(args, shortoptions, longoptions);
	} catch (GetOptException e) {
		writefln(e.msg);
		showhelp();
		return 1;
	}
	
	foreach (getopt_t item; usedoptions) {
		if (item.option == "-h" || item.option == "--help") {
			showhelp();
			return 0;
		} else if (item.option == "-o" || item.option == "--output") {
			outputfilename = item.value;
		} else if (item.option == "-s" || item.option == "--size") {
			fontsize = cast(uint)atoi(item.value);
		} else if (item.option == "-t" || item.option == "--texture-size") {
			texturewidth = textureheight = cast(uint)atoi(item.value);
		} else if (item.option == "-l" || item.option == "--line-spacing") {
			linespacing = atof(item.value);
		} else if (item.option == "-x" || item.option == "--index") {
			fontindex = atoi(item.value);
		} else if (item.option == "-b" || item.option == "--bold") {
			fontbold = true;
		} else if (item.option == "-i" || item.option == "--italic") {
			fontitalic = true;
		} else if (item.option == "-u" || item.option == "--offset-left") {
			offset_left = cast(int)atoi(item.value);
		} else if (item.option == "-v" || item.option == "--offset-top") {
			offset_top = cast(int)atoi(item.value);
		} else {
			writefln("Warning: Unrecognised option '%s'", item.option);
		}
	}
	
	// Input filename is required...
	if (!inputfilename.length) {
		writefln("Error: Please specify a font file to read.");
		showhelp();
		return 1;
	}
	// ...and defaults to a TTF file
	inputfilename = defaultExt(inputfilename, "ttf");
	
	// Output filename defaults to input filename (but with different extensions, and no path)
	if (!outputfilename.length) {
		outputfilename = getBaseName(inputfilename);
	}
	char[] outputcelfilename = addExt(outputfilename, "cel");
	outputfilename = addExt(outputfilename, "bmp");
	
	//// Register missing proc callback with Derelict
	Derelict_SetMissingProcCallback(cast(MissingProcCallback)&isSharedProcUnused);
	
	//// Attempt to open font
	DerelictSDL.load();
	DerelictSDLttf.load();
	if(TTF_Init()<0) {
    	writefln("Failed to initialise font library: "~.toString(TTF_GetError()));
    	return 2;
	}
	
	TTF_Font* font = TTF_OpenFontIndex(.toStringz(inputfilename), fontsize, fontindex);
	
	if (!font) {
		writefln("Failed to open %s (size %s): %s", inputfilename, fontsize, .toString(TTF_GetError()));
		if (fontindex>0) {
			writefln("Perhaps the file does not contain this many fonts; try a lower-numbered font\nindex.");
		}
		TTF_Quit();
		return 3;
	}
	
	debug {
		writefln("Input file: %s, CEL file: %s, BMP file: %s, font size: %d",
			inputfilename, outputcelfilename, outputfilename, fontsize);
	}
	
	//// Set font style
	int style=0;
	if (fontbold) style |= TTF_STYLE_BOLD;
	if (fontitalic) style |= TTF_STYLE_ITALIC;
	if (fontunderline) style |= TTF_STYLE_UNDERLINE;
	if (style != TTF_GetFontStyle(font)) {
		TTF_SetFontStyle(font, style);
	}
	
	//// Write CEL file
	int[] celdata;
	
	// First number is a version number or something similar - 0 apparently works well
	celdata ~= 0;
	
	// Second number is height of a line
	int lineheight = cast(int)(TTF_FontLineSkip(font) * linespacing);
	celdata ~= lineheight;
	
	// Now we calculate character data for the CEL file.
	// The CEL file lists characters encoded in Latin-1 format (0 to 0xFF).
	// Thankfully, Latin-1 is exactly what SDL_ttf accepts! So we just
	// loop from 0 to 0xFF (255) and ask SDL_ttf how wide the character is.
	int charwidth, xmin, xmax;
	int linewidth=0;
	char[2] c; // This array will hold the character number of the glyph...
	c[1] = '\0'; // ...and a null terminator
	
	for (c[0] = 0; c[0] < 255; c[0] = c[0] + 1) {
		if ((c[0] < 32) || (TTF_GlyphMetrics(font, cast(ushort)(c[0]), &xmin, &xmax, null, null, &charwidth) < 0)) {
			// Failed; couldn't render that glyph, so put in a 0 entry
			// (meaning skip it)
			// Note that values below 32 (space character) are automatically skipped;
			// even if they're in the font, we don't want them.
			celdata ~= 0;
			debug writefln("Skipped character %s (%d)", ((c[0]<128 && c[0]!=7) ? c[0] : ' '), c[0]);
		} else {
			
			if (xmax > charwidth) charwidth = xmax;
			debug if (xmin != 0) writefln("Character %d has xmin %d, xmax %d", c[0], xmin, xmax);
			if (xmin < 0) charwidth -= xmin;
			
			debug writefln("Character %s (%d) has width %s", ((c[0]<128 && c[0]!=7) ? c[0] : ' '), c[0], charwidth);
			
			// The width of this character is in charwidth
			if ((linewidth+charwidth) > texturewidth) {
				// Too long! Wrap to next line.
				celdata ~= -1; // Add a line break
				linewidth = 0; // The next line has no width yet
				debug writefln("Wrapping before character %d", c[0]);
			}
			
			// Add character to line
			linewidth += charwidth;
			celdata ~= charwidth;
		}
	}
	// Always skip character 255
	celdata ~= 0;
	
	// Write to file
	char[] celtext;
	for (int i=0; i<celdata.length; i++) {
		if (i!=0) celtext ~= ", ";
		celtext ~= .toString(celdata[i]);
	}
	write(outputcelfilename, celtext);
	
	//// Write BMP file
	// Initialise SDL because we're going to be using some of its functions
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		writefln("Failed to initialise SDL: ", .toString(SDL_GetError()));
		return 4;
	}
	
	// Call SDL_SetVideoMode() to create a dummy video surface - we don't use this,
	// but it needs to be created to keep SDL happy. The size is 1x1 and SDL_NOFRAME
	// is set to prevent people from noticing the window flashing up.
	SDL_Surface* videosurface = SDL_SetVideoMode(1, 1, 0, SDL_SWSURFACE|SDL_NOFRAME);
	if (!videosurface) {
		writefln("Failed to create video surface: ", .toString(SDL_GetError()));
		return 5;
	}
	
	// Red/green/blue masks to give to SDL. Values are dependent on machine byte ordering.
	static if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
	    uint rmask = 0xff000000;
	    uint gmask = 0x00ff0000;
	    uint bmask = 0x0000ff00;
	} else {
	    uint rmask = 0x000000ff;
	    uint gmask = 0x0000ff00;
	    uint bmask = 0x00ff0000;
	}
	
	// Create a surface to write to
	SDL_Surface* bitmap = SDL_CreateRGBSurface(SDL_SWSURFACE, texturewidth, textureheight, 32, rmask, gmask, bmask, 0);
	if (!bitmap) {
		writefln("CreateRGBSurface failed: ", .toString(SDL_GetError()));
		return 6;
	}
	
	// Write glyphs to the surface
	short destx=0, desty=0;
	
	SDL_Color color, bg;
	color.r = color.g = color.b = 255;
	bg.r = bg.g = bg.b = 0;
	
	SDL_Surface* glyph;
	int i = -1;
	
	foreach (int cel; celdata[2..celdata.length]) {
		// Increment glyph number
		i++;
		
		if (cel == -1) {
			// Skip line
			destx = 0;
			desty += lineheight;
			
			// Undo above increment
			i--;
			
		} else if (cel > 0) {
			// Get position to render glyph at
			int minx=0, maxy=0, advance=0;
			if (TTF_GlyphMetrics(font, cast(ushort)i, &minx, null, null, &maxy, &advance) <0 ) {
				writefln("Warning: Failed to get metrics of character %d", i);
			}
			
			debug writefln("Character %d has x and y: %d, %d", i, minx, maxy);
			
			// Render this glyph
			glyph = TTF_RenderGlyph_Shaded(font, cast(ushort)i, color, bg);
			SDL_Rect destrect;
			destrect.x = cast(short)(destx + offset_left);
			destrect.y = cast(short)(desty - maxy + TTF_FontAscent(font) + TTF_FontDescent(font)/2 + offset_top);
			if (!glyph || SDL_BlitSurface(glyph, null, bitmap, &destrect)<0) {
				writefln("Warning: Blit of character %d failed", i);
			}
			
			// Move across for next glyph
			destx += cel;
		}
	}
	
	// Save surface
	SDL_SaveBMP(bitmap, .toStringz(outputfilename));
	
	// Clean up
	TTF_Quit();
	SDL_Quit();
	
	return 0;
}

bool isSharedProcUnused(char[] libName, char[] procName) {
	if (procName != "SDL_GetKeyRepeat") return false;
	else return true;
}

void showhelp() {
	writefln("\nFontTaffer v%s by Crispy", progVersion,"\n");
	writefln("Usage: fonttaff truetype.ttf [options]");
	writefln("       fonttaff raster.fon [options]");
	writefln("    -h, --help                 Displays this message");
	writefln("    -s, --size=FONTSIZE        Font size to use (default 90)");
	writefln("    -o, --output=FILENAME      Output filename (optional)");
	writefln("    -t, --texture-size=PIXELS  Size of output texture (default 512)");
	writefln("    -l, --line-spacing=FACTOR  Line spacing (default 1.5)");
	writefln("    -b, --bold                 Render the font in bold type");
	writefln("    -i, --italic               Render the font in italics");
	writefln("    -u, --offset-left=PIXELS   Shift all glyphs PIXELS pixels to the right");
	writefln("                               Use this to avoid minor cutoffs (default 0)");
	writefln("    -v, --offset-top=PIXELS    Shift all glyphs PIXELS pixels downwards");
	writefln("                               Use this to avoid minor cutoffs (default 0)");
	writefln("    -x, --index=FONTINDEX      Font index (default 0). You will");
	writefln("                               usually not have to change this.");
	writefln("Note: No wildcards are allowed in filenames.");
	writefln("Note: The --offset-left and --offset-right options do not modify the .cel");
	writefln("      file, just the bitmap. Use small values to avoid creating problems!");
	writefln("Thanks to New Horizon and everyone else at TTLG!");
	writefln("This program is based in part on the work of the FreeType team\n(http://www.freetype.org)");
}